// Challenge 3, Jonas Dedden

#include "float.h"
#include "math.h"

#define sqrt2 0.7071067811865475
#define sinPi8 0.3826834323650898
#define cosPi8 0.9238795325112868

const double nodes9[] = {-1.0, -cosPi8, -sqrt2, -sinPi8, 0, sinPi8, sqrt2, cosPi8, 1};

const double weights5[] = {
        0.0666666666666666,
        0.5333333333333334,
        0.8000000000000000,
        0.5333333333333334,
        0.0666666666666666
};

const double weights9[] = {
        0.0158730158730158,
        0.1462186492160182,
        0.2793650793650794,
        0.3617178587204898,
        0.3936507936507936,
        0.3617178587204898,
        0.2793650793650794,
        0.1462186492160182,
        0.0158730158730158
};

double integrateN(double a, double b, double t, double (*f)(double), double first, double last){
    t = fmax(DBL_EPSILON, t); // Upper limit for accuracy is the machine precision
    double ba = (b-a)/2;
    if ((b-a) <= t/2)
        return ba*(first + last);
    else {
        double ba2 = (a+b)/2;

        double fValues[9];

        fValues[0] = first;
        fValues[8] = last;

        double k9 = 0;
        k9 += weights9[0]*fValues[0] + weights9[8]*fValues[8];
        for (int i = 1; i<8; i++){
            fValues[i] = (*f)(ba2 + ba*nodes9[i]);
            k9 += weights9[i]*fValues[i];
        }
        k9 *= ba;

        double k5 = 0;
        for (int i = 0; i<5; i++)
            k5 += weights5[i]*fValues[i*2];
        k5 *= ba;

        if (fabs(k9 - k5) <= t)
            return k9;
        else
            return integrateN(a, ba2, t/2, f, first, fValues[4]) + integrateN(ba2, b, t/2, f, fValues[4], last);
    }
}

double integrate(double a, double b, double t, double (*f)(double)){
    return integrateN(a, b, t, f, (*f)(a), (*f)(b));
}
